#ifndef ImageDecoderWiiU_h
#define ImageDecoderWiiU_h

namespace WebCore {
	
class ImageWiiU
{
	private:
		int m_refCount;
		void* m_textureP;
		int m_format;
		bool m_dither;
		bool m_isDirect;
		bool m_isDirty;
		bool m_isFinished;

		void* createTexture(int width, int height, bool hasAlpha, void *sourcePixels, bool noCopy, bool allowBC1, bool allowFormats);
		ImageWiiU(int width, int height, bool hasAlpha, void *sourcePixels, bool noCopy, bool allowBC1, bool allowFormats);
		ImageWiiU(void *textureDataP, bool isDirect);
		ImageWiiU(ImageWiiU *originalP);
		~ImageWiiU(void);
	public:
		static ImageWiiU* create(void *textureDataP, bool isDirect=true);	// creates a "direct" texture that cannot be freed (if isDirect==true)
		static ImageWiiU* create(int width, int height, bool hasAlpha=true, void *sourcePixels=NULL, bool noCopy=true, bool allowBC1=false, bool allowFormats=false);
		static ImageWiiU* create(ImageWiiU *originalP);
		void ref(void);
		void unref(void);
		void writeNativeImagePixelsOpaque(int y, int width, unsigned char *sourcePixelsP, int step);
		bool writeNativeImagePixelsAlpha(int y, int width, unsigned char *sourcePixelsP, int step, bool preMultiplyAlpha);
		void writeNativeImagePixels(int y, int width, unsigned long *sourcePixelsP);
		bool writeNativeImagePixel(int x, int y, unsigned long argb);
		bool writeNativeImagePixel(int x, int y, unsigned char r, unsigned char g, unsigned char b, unsigned char a, bool preMultiplyAlpha);
		void setDither(bool enabled)	{ m_dither = enabled; }
		void copyRowNTimes(int startX, int endX, int startY, int endY);
		void invalidateData(void);
		void zeroData(void);
		void* getTexture(void);
		bool isDirectTexture(void) { return m_isDirect; }
};

} // namespace WebCore
#endif
