/*
 * Copyright (C) 2007 Holger Hans Peter Freyther
 * Copyright (C) 2007, 2008 Alp Toker <alp@atoker.com>
 * Copyright (c) 2012 NINTENDO.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef WEB_VIEW_WIIU_H
#define WEB_VIEW_WIIU_H

#include <wtf/Platform.h> // PLATFORM(...)

#if defined(WIN32) && !PLATFORM(WIIU_DEV)
#if WEBKIT_IMPLEMENTATION
    #define WEBKIT_API __declspec(dllexport)
#else
    #define WEBKIT_API __declspec(dllimport)
#endif
#else
#define WEBKIT_API __attribute__((visibility("default")))
#endif

//#define USE_OLD_WEBVIEW_RESTART 0

#include <wtf/text/CString.h>
#include <WebKit/wiiu/Api/wk_event.h>
#include <string.h>

typedef enum {
    WEBVIEW_LOAD_STARTED,
    WEBVIEW_LOAD_COMPLETE,
    WEBVIEW_STATUS,
    WEBVIEW_TITLE_CHANGED,
} WebViewState_t;

typedef struct {
    int x, y, width, height;
} WebViewRect_t;

struct WebViewSize_t {
	int width;
	int height;
};

//APW 9/28/12 - I added INIT, UNINT, and READY to use with the SystemEventDispatcher.  If this is used somewhere else, I can create a new enum
typedef enum {
	WEBVIEW_EVENT_UNINIT = 0,
	WEBVIEW_EVENT_INIT,
	WEBVIEW_EVENT_READY,
	WEBVIEW_EVENT_LOAD,
}WebViewEventState_t;

typedef void * WebViewHandle_t;

// Adapted from WebViewManager::DisplayType -TS-NSA 2012-03-12
enum WebViewType_t {
	WEBVIEW_TV,
	WEBVIEW_GAMEPAD,
	WEBVIEW_NONE
};

// Location to view Inspector front-end
enum WebViewInspectorLocation_t {
    WEBVIEW_INSPECTOR_NONE = 0,
    WEBVIEW_INSPECTOR_SPLIT_TV,
    WEBVIEW_INSPECTOR_GAMEPAD0,
    WEBVIEW_INSPECTOR_GAMEPAD1,
    WEBVIEW_INSPECTOR_GAMEPAD2,
    WEBVIEW_INSPECTOR_GAMEPAD3,
    WEBVIEW_INSPECTOR_REMOTE,
};

// this maps to the enumeration in Cursor.h
enum WebViewCursorType {
    Pointer,
    Cross,
    Hand,
    IBeam,
    Wait,
    Help,
    EastResize,
    NorthResize,
    NorthEastResize,
    NorthWestResize,
    SouthResize,
    SouthEastResize,
    SouthWestResize,
    WestResize,
    NorthSouthResize,
    EastWestResize,
    NorthEastSouthWestResize,
    NorthWestSouthEastResize,
    ColumnResize,
    RowResize,
    MiddlePanning,
    EastPanning,
    NorthPanning,
    NorthEastPanning,
    NorthWestPanning,
    SouthPanning,
    SouthEastPanning,
    SouthWestPanning,
    WestPanning,
    Move,
    VerticalText,
    Cell,
    ContextMenu,
    Alias,
    Progress,
    NoDrop,
    Copy,
    None,
    NotAllowed,
    ZoomIn,
    ZoomOut,
    Grab,
    Grabbing,
    Custom
};

// Valid places that can block and unblock webview restart events
enum WebViewRestartBlockSource {
	WEBVIEW_RESTART_BLOCKSOURCE_WEBVIEW_TV  = 0x01,
	WEBVIEW_RESTART_BLOCKSOURCE_WEBVIEW_GP0 = 0x02,
};

enum WebViewScrollStyle {
	SCROLL_HORIZONTAL,
	SCROLL_VERTICAL,
	SCROLL_FREE,
	SCROLL_NONE
};

struct  WebViewCallbacks_t {
    void (*status)(WebViewState_t type, const char *data, void *user_data);
    void (*error)(const char *url, const char *error, void *user_data);
    void (*progress)(int percent, void *user_data);
    WebViewHandle_t (*open_window)(const char *url, void *user_data);
	bool (*authenticate)(const char *url, char **user, char **pass, void *user_data);

    void (*console_message)(const wchar_t *message, std::size_t message_len, int line_number, const wchar_t *sourceID, void *user_data);
    void (*js_alert)(const wchar_t * msg, std::size_t msg_len, void *user_data);
    bool (*js_confirm)(const wchar_t * msg, std::size_t msg_len, void *user_data);
    bool (*js_prompt)(const wchar_t *message, std::size_t message_len, const wchar_t *default_result, std::size_t default_result_len, wchar_t **result, void *user_data);
    void (*set_cursor)(WebViewCursorType type, void *cursor);
};

typedef void (WebViewPaintCB_t)(const void *texture, WebViewHandle_t webview, void *user_data);

struct WebViewSettings_t {
    bool			LoadsImagesAutomatically;
	// views
	WebViewSize_t	tvSize;
	WebViewSize_t	gpSize;
    // fonts
    int				DefaultFixedFontSize;
    int				DefaultFontSize;
    const char *	SerifFontFamily;
    const char *	FixedFontFamily;
    const char *	SansSerifFontFamily;
    const char *	StandardFontFamily;
    // javascript
    bool			JavaScriptEnabled;
    bool			JavaScriptCanOpenWindowsAutomatically;
    // storage
    bool			LocalStorageEnabled;
    bool			OfflineWebApplicationCacheEnabled;
    const char *	LocalStorageDatabasePath;
    unsigned		SessionStorageQuota;
    // cache
    size_t			minDeadCapacity;
    size_t			maxDeadCapacity;
    size_t			capacity;
    double			deletionInterval;
    int				pageCacheSize;
    bool			developerExtras;
    bool            bExtendedDisplay;
	//Required RPL list
	std::string		requiredRPLList;
};

WEBKIT_API void DispatchSystemEvent(const char *event_id, void *data);
WEBKIT_API void DispatchInputEvents(void);
WEBKIT_API void DispatchTestInputEvents(void);
WEBKIT_API void WebKitOutOfMemory(void (*callback)(void));
WEBKIT_API bool WebKitReadyForReset(void);

// Webview interface.
WEBKIT_API bool WebViewsInitialized(void);
WEBKIT_API void WebViewsCreate(WebViewCallbacks_t *callbacks, WebViewSettings_t *settings, unsigned int gamepad_web_view_count, WebViewInspectorLocation_t location);
WEBKIT_API void WebViewsDestroy(void);
WEBKIT_API WebViewHandle_t WebViewTV(void);
WEBKIT_API WebViewHandle_t WebViewGamePad(unsigned int index);
WEBKIT_API WebViewHandle_t WebViewInspector(void);
WEBKIT_API WebViewHandle_t WebViewAny(void);
WEBKIT_API WebViewType_t WebViewType(WebViewHandle_t webview);
WEBKIT_API int NumWebViews(void);
WEBKIT_API int NumGamePadWebViews(void);
WEBKIT_API bool WebViewRendersToTV(WebViewHandle_t view);
WEBKIT_API int  WebViewRendersToLocation(WebViewHandle_t view);
WEBKIT_API void WebViewLoad(WebViewHandle_t view, const char *data);
WEBKIT_API void WebViewSetVisible(WebViewHandle_t view, bool visible);
WEBKIT_API bool WebViewIsVisible(WebViewHandle_t view);
WEBKIT_API unsigned WebViewGetWidth(WebViewHandle_t view);
WEBKIT_API unsigned WebViewGetHeight(WebViewHandle_t view);
WEBKIT_API void *WebViewPage(WebViewHandle_t view);
WEBKIT_API WebViewHandle_t WebViewCreate(WebViewSize_t *size, void *surface, WebViewCallbacks_t *cbs, WebViewSettings_t *settings, bool noSetup, void *user_data, WebViewHandle_t inspectorview, bool inactiveDueToExtDisplay = false);
WEBKIT_API void WebViewDestroy(WebViewHandle_t view, bool bOnWebkitThread);
WEBKIT_API void WebViewSaveScreen(WebViewHandle_t view, const char *filename, bool extendedSaveGamePad = false);
WEBKIT_API void *WebViewCaptureScreen(WebViewHandle_t view, int dstWidth, int dstHeight, int srcX = 0, int srcY = 0, int srcWidth = -1, int srcHeight = -1);
WEBKIT_API void WebViewSetUA(WebViewHandle_t view, const char *agent);
WEBKIT_API void WebViewSetCustomHeader(WebViewHandle_t view, const char *header);
WEBKIT_API int WebInspectorPort(void);
WEBKIT_API void WebViewSetTouchScrollEnabled(WebViewHandle_t view, bool enabled);
WEBKIT_API void WebViewSetScrollbarsEnabled(WebViewHandle_t view, bool enabled);
WEBKIT_API void WebViewSetTouchEventsEnabled(WebViewHandle_t view, bool enabled);
WEBKIT_API void WebViewSetStartVideoInFullscreenEnabled(WebViewHandle_t view, bool enabled);

enum WaitForEventUrlEvents {
	WAITEVENTURL_INSPECTOR_FRONTEND = 0x01,
	WAITEVENTURL_HOST_RECONNECT     = 0x02,
	WAITEVENTURL_CUSTOM_FONTS       = 0x04,
};
void WebViewSetWaitUrlEvent(int waitEvent);
void WebViewSetWaitForEventUrl(WebViewHandle_t view, const char *url, int waitEvents);
void WebViewInitEventUrls();
void WebViewClearEventUrls();
CString WebInspectorFrontendUrl(void);
void WebKitEnableDraw(bool enable);
void WebKitSetRestartWebViews(bool restart, float fadeTimeMs = 0, const char * vXMLConfigData = 0, const unsigned int vXMLConfigLength = 0);
bool WebKitRestartWebViews(void);
float WebKitGetRestartWebViewsFadeOutTime(void);
void WebKitIncrementRestartCount(void);
int WebKitGetRestartCount(void);
void WebViewBlockRestartWebViews(bool block, int blockSource);
WebViewCallbacks_t *WebViewGetCallbacks(void);
WebViewSettings_t *WebViewGetSettings(void);
WebViewInspectorLocation_t WebViewGetInspectorLocation(void);
CString WebViewGetURL(WebViewHandle_t view);


typedef void  (WebKitPaintCB_t)(const void *texture, WebViewHandle_t webview, void *user_data);
typedef void* (WebKitReallocCB_t)(void *pAddress, size_t newByteSize, signed heap);
typedef void  (WebKitAcquireGpuFgResourcesCB_t)(bool bAcquire);

struct WebKitCallbacks_t {
	WebKitPaintCB_t					(*m_paintCB);
	WebKitReallocCB_t				(*m_reallocCB);
	WebKitAcquireGpuFgResourcesCB_t	(*m_acquireGpuFgResourcesCB);
};

struct WebKitCustomRenderData {
	WebKitCustomRenderData *nextP, *prevP;	// for linked list
    virtual bool DoPreAdd(void) {return true;}
	virtual bool DoRender(void *colorBufferP, long width, long height, bool isGamePad) = 0;
	virtual void DoneRender(void) = 0;
};

WEBKIT_API bool WebKitStartup(const WebKitCallbacks_t *pCallbacks, int prio, int core);
WEBKIT_API void WebKitShutdown(const bool vDeleteInstance);
WEBKIT_API void WebKitTriggerShutdown(void);
WEBKIT_API bool WebKitDone(void);
WEBKIT_API bool WebKitShuttingDown(void);
WEBKIT_API void WebKitPause(bool bPauseMedia = true);
WEBKIT_API void WebKitResume(void);
#if !PLATFORM(WIIU)
WEBKIT_API bool WebKitIsPaused(void);
WEBKIT_API void WebKitWorkPing(void);
#endif
WEBKIT_API void WebKitSignalReadyForShutdown(void);
WEBKIT_API void WebKitWaitForShutdown(void);
WEBKIT_API void WebKitQuit(void);
WEBKIT_API bool WebKitProcessEvents(bool fromEventLoop = false);

WEBKIT_API bool WebKitIsBackground(void);

WEBKIT_API void WebKitSetProxyInfo(const char *host, unsigned long port, WebKitProxyType type, const char *username, const char *password);
WEBKIT_API void WebKitSetCertificate(const char *cert);

WEBKIT_API void WebKitAddCustomRenderingCommand(WebKitCustomRenderData *dataP);
WEBKIT_API void WebKitRemoveCustomRenderingCommand(WebKitCustomRenderData *dataP);
WEBKIT_API void WebKitRenderCustomEventsHold(void);
WEBKIT_API void WebKitRenderCustomEventsRelease(void);

WEBKIT_API void WebKitRendering(void);
WEBKIT_API void WebKitRenderCustomEvents(void *colorBufferP, long width, long height, bool isGamePad);
WEBKIT_API int WebKitSendEvent(WiiUBaseEventData *data);
WEBKIT_API int WebKitSendEventVSync(WebViewHandle_t view);
WEBKIT_API int WebKitSendEventMouse(int eventType, WebViewHandle_t view, int x, int y, int button, int state, int clickCount = 0, long timestamp = 0);
WEBKIT_API int WebKitSendEventTouch(int eventType, WebViewHandle_t view, int x, int y, int button, int state, int clickCount = 0, long timestamp = 0);
WEBKIT_API void WebKitSendEventAutoScrollStop(WebViewHandle_t view);
WEBKIT_API int WebKitSendEventInputUpdate();
WEBKIT_API void WebKitSendEventUSBKeyboard(bool attached);
WEBKIT_API int WebKitSendEventDelayed(WiiUBaseEventData *data, int delayMs = 0);

WEBKIT_API void *WebKitRealloc(void *pAddress, size_t newByteSize);

//CRANKE-ME adding Web Inspector
WEBKIT_API void   WebViewEnableInspector(WebViewHandle_t view, bool enable);
WEBKIT_API void   WebViewEnableRemoteInspector(int port, std::string & frontendUrl, std::string & hostUrl);

//Restart support
WEBKIT_API void WebKitResetGlobals();
WEBKIT_API void WebKitDestroyMediaPlayerGlobals();

//For reloading the Config XML
WEBKIT_API const char * WebKitGetRestartWebViewsConfigXML();
WEBKIT_API const unsigned int WebKitGetRestartWebViewsConfigXMLLength();

WEBKIT_API void WebKitForceGC();

WEBKIT_API void WebKitSetEnhancedUserInterfaceAccessibility();
WEBKIT_API void WebKitResetEnhancedUserInterfaceAccessibility();

WEBKIT_API void WebViewSetSendInspectorConnectedCallback(void (*fn)());
WEBKIT_API void WebViewSetSendDashboardSpecificMessage(void (*fn)(unsigned int));

//Screenshot completion notice
WEBKIT_API void WebViewSetScreenshotCompleteCallback(void (*fn)(const char *, bool));
WEBKIT_API void WebViewAddScreenshotCompleteCallback(void (*fn)(const char *, bool));
WEBKIT_API void WebViewRemoveScreenshotCompleteCallback(void (*fn)(const char *, bool));
void QueuedScreenshotCompleteCallback(const char*, bool);

//Screenshot progress update
WEBKIT_API void WebViewSetScreenshotProgressCallback(bool (*fn)(const char *, void *, int), int progressFrequency);

//Stop any queued screenshots
WEBKIT_API void WebViewAbortScreenshots();

WEBKIT_API bool WebViewUseOldRestart();
WEBKIT_API bool WebViewSetUseOldRestart(const bool vUseOldRestart);

WEBKIT_API void WebKitDrop(void);
WEBKIT_API bool WebKitGetPanicExit(void);

#if !PLATFORM(WIIU)
WEBKIT_API void SetAssetsDir(const char *set);
WEBKIT_API void SetContentDir(const char *set);
WEBKIT_API void SetSaveDir(const char *set);
WEBKIT_API void SetSaveCommonDir(const char *set);

WEBKIT_API const char *GetContentDir(void);
WEBKIT_API const char *GetSaveDir(void);
WEBKIT_API const char *GetSaveCommonDir(void);
WEBKIT_API const char *GetSaveCommonTempDir(void);
#endif

#endif
