/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __WK_EVENT_H
#define __WK_EVENT_H

#include <WiiUEvent.h>
#include <RefPtr.h>
#include <StorageSyncCallback.h>

#include <stdlib.h>
#include <string.h>

#if OS(WINDOWS)
/* FIXME: CLEANUP -- because mingw-gcc has its own headers for e.g. <sys/...>,
    we have to avoid adding 'libshim/include' as a -I path, and hence need to
    spell out fuller #include paths for headers that *are* within libshim.
    There must be a better way to organize this.
   -TS-NSA 2012-08-16 */
#include <libshim/include/pthread.h>
#else
#include <pthread.h>
#endif

#include <string>

struct WebViewCallbacks_t;
struct WebViewSettings_t;
struct WebViewSize_t;
class WebView;

// -------------------------------------------------------------------------------------------

//! User event types (not used in WebCore)
enum {
	WIIU_FOCUS_GAINED = WIIU_USER_EVENT_START,
	WIIU_FOCUS_LOST,
	WIIU_NEXT_PAGE,
	WIIU_PREV_PAGE,
	WIIU_REFRESH_PAGE,
	WIIU_STOP_LOAD,
	WIIU_LOAD_PAGE,
	WIIU_CLEAR_PAGE,
	WIIU_SCROLL_UP,
	WIIU_SCROLL_DOWN,
	WIIU_SCROLL_LEFT,
	WIIU_SCROLL_RIGHT,
	WIIU_LOAD_URL,
	WIIU_SCREENSHOT,
	WIIU_PAUSE_MEDIA,

	WIIU_CREATE_VIEW,
	WIIU_DESTROY_VIEW,
	WIIU_SET_USER_AGENT,
	WIIU_SET_CUSTOM_HEADER,
	WIIU_SET_PROXY_INFO,
	WIIU_SET_CERTIFICATE,
	
	// to enable sending of IO events from a callback
	WIIU_IO_EVENT,
	// to enable sending of AOC events from a callback
	WIIU_AOC_EVENT,
	// account independent service token
	WIIU_ACCOUNT_SERVICE_TOKEN,
	// network connection/disconnection events
	WIIU_NETWORK_EVENT,
    // when software keyboard input has finished
    WIIU_SWKBD_EVENT,
    // when software keyboard animation has started or finished
    WIIU_SWKBD_ANIMATION_EVENT,
	// distribute input events to JS extentions
	WIIU_INPUT_UPDATE,
	// set the contents of a text field in one fell swoop
	WIIU_SET_TEXT_FIELD_EVENT,
	// unfocus the currently focused element
	WIIU_BLUR_CURRENT,
	// stop the autoscroller
	WIIU_STOP_AUTOSCROLLER,
	// break out of webkit mainloop
	WIIU_INTERRUPT_WEBVIEW,
	// stop processing events
	WIIU_STOP_EVENT_LOOP,

    // browser has just synchronized to vsync
    WIIU_DISPLAY_VSYNC,
	// accept remote inspector connections
    WIIU_INSPECTOR_SERVER_ACCEPT,
    // Remote inspector server start. Enables listen & set Frontend URL.
    WIIU_INSPECTOR_SERVER_START,
    // Remote inspector server stop.
    WIIU_INSPECTOR_SERVER_STOP,
    //Inspector Enable
    WIIU_INSPECTOR_ENABLE,
	// distribute error events to JS extentions
	WIIU_ERROR_UPDATE,
	// distribute events to JS extentions
	WIIU_EXTENSION_EVENT,
	WIIU_CUSTOM_EVENT,
	//Destroy all of the current webviews and then restart them again with the last page that they had. 
	WIIU_RESTART_WEBVIEWS,
	WIIU_BG_NOTIFICATION,
	WIIU_FG_NOTIFICATION,
	WIIU_SHUTDOWN_NOTIFICATION,

	WIIU_RESTART_FINISHED,
	WIIU_PREPARE_FOR_RESTART,

	// The usb keyboard has attached or detached
	WIIU_USB_KEYBOARD_EVENT,

	// Inspector frontend is connected
	WIIU_INSPECTOR_FRONTEND_CONNECTED,

	// Dashboard is reconnected after a restart
	WIIU_HOST_RECONNECTED,

	// Custom fonts have all been loaded over running client connection
	WIIU_CUSTOM_FONTS_LOADED,

	// Check if any urls are ready to be loaded
	WIIU_TRIGGER_EVENT_URLS,

	WIIU_LIGHT_GC,
	WIIU_FORCE_GC,

	WIIU_STORAGE_SYNC,

	//HIO Connection messages
	WIIU_HIO_NEW_CONNECTION,
	WIIU_HIO_NO_CONNECTIONS,
    WIIU_PLAY_PAUSE_MEDIA,
    WIIU_ADVANCE_MEDIA, 
    WIIU_SEEK_START_MEDIA, 
    WIIU_SEEK_UPDATE_MEDIA,
    WIIU_SEEK_END_MEDIA,
    WIIU_CURRENT_TIME_MEDIA,
    WIIU_QUIT_MEDIA,
    WIIU_PLAYBACK_RATE_MEDIA
};

enum {
	WIIU_BUTTON_NONE = -1,
	WIIU_LEFT,
	WIIU_MIDDLE,
	WIIU_RIGHT,
};

enum {
	WIIU_BUTTON_DOWN,
	WIIU_BUTTON_UP,
	WIIU_KEYPRESS_DOWN,
	WIIU_KEYPRESS_UP
};

enum WebKitProxyType {
	WEBKITPROXY_HTTP,
	WEBKITPROXY_SOCKS4,
	WEBKITPROXY_SOCKS4A,
	WEBKITPROXY_SOCKS5,
	WEBKITPROXY_SOCKS5HOSTNAME
};

enum {
    WIIU_SWKBD_OK,
    WIIU_SWKBD_CANCEL,
    WIIU_SWKBD_KEY,
    WIIU_SWKBD_START,
    WIIU_SWKBD_END,
    WIIU_SWKBD_SET_TEXT,
};

class WiiUPrepareForRestartEvent : public WiiUBaseEventData
{
public:
	WiiUPrepareForRestartEvent(pthread_cond_t * vSignal, pthread_mutex_t * vMutex, bool * vProcessed) : m_pSignal(vSignal), m_pMutex(vMutex), m_pProcessed(vProcessed){}
	virtual WiiUEventType_t	type() const	{ return WIIU_PREPARE_FOR_RESTART; }

public:
	volatile bool * m_pProcessed;
	pthread_cond_t * m_pSignal;
	pthread_mutex_t * m_pMutex;
};

class WiiURestartFinishedEvent : public WiiUBaseEventData
{
public:
	WiiURestartFinishedEvent(pthread_cond_t * vSignal, pthread_mutex_t * vMutex) : m_pSignal(vSignal), m_pMutex(vMutex){}
	virtual WiiUEventType_t	type() const	{ return WIIU_RESTART_FINISHED; }

public:
	pthread_cond_t * m_pSignal;
	pthread_mutex_t * m_pMutex;
};

class WiiUScrollEvent : public WiiUBaseEventData
{
public:
	WiiUScrollEvent(WiiUEventType_t _evType, WebViewHandle_t _webview, float _scale = 1.0f) : webview(_webview), evType(_evType), scale(_scale) {}
	
	WebViewHandle_t		webview;
	int					evType;
	float				scale;

	virtual WiiUEventType_t	type() const	{ return (WiiUEventType_t)evType; }
};

class WiiUFocusEvent : public WiiUBaseEventData
{
public:
	WiiUFocusEvent(WiiUEventType_t _evType, WebViewHandle_t _webview) : webview(_webview),evType(_evType) {}
	
	WebViewHandle_t		webview;
	int					evType;

	virtual WiiUEventType_t	type() const	{ return (WiiUEventType_t)evType; }
};

class WiiUNavEvent : public WiiUBaseEventData
{
public:
	WiiUNavEvent(WiiUEventType_t _evType, WebViewHandle_t _webview) : webview(_webview), evType(_evType) {}
	WiiUNavEvent(WebViewHandle_t _webview, const std::string & _url) : url(_url), webview(_webview), evType(WIIU_LOAD_PAGE) {}
	~WiiUNavEvent() {};
	
	std::string		url;
	WebViewHandle_t	webview;
	int				evType;

	virtual WiiUEventType_t	type() const	{ return (WiiUEventType_t)evType; }
};

class WiiULoadURLEvent : public WiiUBaseEventData
{
public:
	WiiULoadURLEvent(WebViewHandle_t _webview, const std::string & _url) : url(_url), webview(_webview) {}
	~WiiULoadURLEvent() {}
	
	std::string		url;
	WebViewHandle_t	webview;

	virtual WiiUEventType_t	type() const	{ return WIIU_LOAD_URL; }
};

class WiiUStopAutoScrollerEvent : public WiiUBaseEventData
{
public:
	WiiUStopAutoScrollerEvent(WebViewHandle_t _webview) : webview(_webview) {}
	~WiiUStopAutoScrollerEvent() {}

	WebViewHandle_t	webview;

	virtual WiiUEventType_t	type() const	{ return WIIU_STOP_AUTOSCROLLER; }
};

class WiiUScreenshotEvent : public WiiUBaseEventData
{
public:
	WiiUScreenshotEvent(WebViewHandle_t _webview, const std::string & _filename) : filename(_filename), webview(_webview) {}
	~WiiUScreenshotEvent() {}
	
	std::string		filename;
	WebViewHandle_t	webview;

	virtual WiiUEventType_t	type() const	{ return WIIU_SCREENSHOT; }
};

class WiiUCreateViewEvent : public WiiUBaseEventData
{
public:
	WiiUCreateViewEvent(WebViewSize_t * _size, WebViewCallbacks_t *_cbs, void (*_paintCB)(const void* , WebViewHandle_t, void*),
						WebViewSettings_t *_settings, bool _bNoSetup, void *_userdata, WebView **_webview, pthread_cond_t *_cond, pthread_mutex_t *_mutex, WebView * _inspectorView, bool inactiveDueToExtDisplay)
		: size(_size), cbs(_cbs), paintCB(_paintCB), settings(_settings), bNoSetup(_bNoSetup), userdata(_userdata), webview(_webview), cond(_cond), mutex(_mutex),
        inspectorview(_inspectorView), bInactiveDueToExtDisplay(inactiveDueToExtDisplay) {}

	WebViewSize_t		*size;
	WebViewCallbacks_t	*cbs;
	void			   (*paintCB)(const void*, WebViewHandle_t, void*);
	WebViewSettings_t	*settings;
	bool				bNoSetup;
	void 				*userdata;
	WebView				**webview;
	pthread_cond_t		*cond;
	pthread_mutex_t		*mutex;
    WebView              *inspectorview;
    bool                bInactiveDueToExtDisplay;

    virtual WiiUEventType_t	type() const	{ return WIIU_CREATE_VIEW; }
};

class WiiUDestroyViewEvent : public WiiUBaseEventData
{
public:
	WiiUDestroyViewEvent(WebViewHandle_t _webview, pthread_cond_t *_cond, pthread_mutex_t *_mutex) : webview(_webview), cond(_cond), mutex(_mutex) {}
	
	WebViewHandle_t	webview;
	pthread_cond_t	 *cond;
	pthread_mutex_t *mutex;

	virtual WiiUEventType_t	type() const	{ return WIIU_DESTROY_VIEW; }
};

class WiiUSetUserAgentEvent : public WiiUBaseEventData
{
public:
	WiiUSetUserAgentEvent(WebViewHandle_t _webview, const std::string & _ua) : webview(_webview), ua(_ua) {}
	~WiiUSetUserAgentEvent() {}
	
	WebViewHandle_t	webview;
	std::string		ua;

	virtual WiiUEventType_t	type() const	{ return WIIU_SET_USER_AGENT; }
};

class WiiUSetCustomHeaderEvent : public WiiUBaseEventData
{
public:
	WiiUSetCustomHeaderEvent(WebViewHandle_t _webview, const std::string & _header) : webview(_webview), header(_header) {}
	~WiiUSetCustomHeaderEvent() {}
	
	WebViewHandle_t	webview;
	std::string		header;

	virtual WiiUEventType_t	type() const	{ return WIIU_SET_CUSTOM_HEADER; }
};

class WiiUSetProxyInfoEvent : public WiiUBaseEventData
{
public:
	WiiUSetProxyInfoEvent(const std::string & _username, const std::string & _password, const std::string & _host, unsigned _port, WebKitProxyType _proxyType)
		: username(_username), password(_password), host(_host), port(_port), proxyType(_proxyType) {}
	~WiiUSetProxyInfoEvent() {}

	std::string		username;
	std::string		password;
	std::string		host;
	unsigned				port;
	WebKitProxyType proxyType;

	virtual WiiUEventType_t	type() const	{ return WIIU_SET_PROXY_INFO; }
};

class WiiUSetCertificateEvent : public WiiUBaseEventData
{
public:
	enum {
		TYPE_CERT,
		TYPE_CLIENT_CERT,
		TYPE_CRL
	};

	WiiUSetCertificateEvent(const std::string & _cert)
		: cert(_cert), certType(TYPE_CERT) {}
	~WiiUSetCertificateEvent() {}

	std::string		cert;
	int				certType;

	virtual WiiUEventType_t	type() const	{ return WIIU_SET_CERTIFICATE; }
};

class WiiUInputUpdate : public WiiUBaseEventData
{
public:
    WiiUInputUpdate() {};
    ~WiiUInputUpdate() {};
	virtual WiiUEventType_t	type() const	{ return WIIU_INPUT_UPDATE; }

	virtual bool rejectExclusive() const;
	virtual void postExclusive();
	virtual void releaseExclusive();

	friend void ResetWkEventGlobals();
	static void Startup(void);
private:
   static bool     s_bInQueue;
};

class WiiUUSBKeyboardData : public WiiUBaseEventData {
public:
    WiiUUSBKeyboardData(bool attached) : attached(attached) {};
    ~WiiUUSBKeyboardData(void) {};
	
	virtual WiiUEventType_t	type(void) const { 
		return WIIU_USB_KEYBOARD_EVENT; 
	}

	bool Attached(void) {return attached;}

private:
	bool attached;
};

class WiiUInterruptWebViewEvent : public WiiUBaseEventData
{
public:
	virtual WiiUEventType_t	type() const	{ return WIIU_INTERRUPT_WEBVIEW; }
};

class WiiUStopEventLoopEvent : public WiiUBaseEventData
{
public:
	virtual WiiUEventType_t	type() const	{ return WIIU_STOP_EVENT_LOOP; }
};

class WiiUSoftwareKeyboardEvent : public WiiUBaseEventData
{
public:
    WiiUSoftwareKeyboardEvent(int _evType, const std::wstring & _msg = L"", WebViewHandle_t _webview = NULL) : evType(_evType), msg(_msg), webview(_webview) {}
    ~WiiUSoftwareKeyboardEvent() {}

    int             evType;
    std::wstring     msg;
	WebViewHandle_t webview;

    virtual WiiUEventType_t type() const    { return WIIU_SWKBD_EVENT; }
};

class WiiUSetTextFieldDataEvent : public WiiUBaseEventData
{
public:
    WiiUSetTextFieldDataEvent(void *_pNode, const std::string & _data = "") : pNode(_pNode), data(_data) {}
    ~WiiUSetTextFieldDataEvent() {}

    void        *pNode;
    std::string data;

    virtual WiiUEventType_t type() const    { return WIIU_SET_TEXT_FIELD_EVENT; }
};

class WiiUBlurEvent : public WiiUBaseEventData
{
	WebViewHandle_t m_webview;
public:
	explicit WiiUBlurEvent(WebViewHandle_t view) : m_webview(view) {}
	virtual WiiUEventType_t type() const	{ return WIIU_BLUR_CURRENT; }
	WebViewHandle_t webview() const	{ return m_webview; }
};

// Event for Remote Web Inspector Accept
class WiiUWebInspectorServerAcceptEvent : public WiiUBaseEventData
{
public:
    WiiUWebInspectorServerAcceptEvent(int fd)
		: connectionFd(fd)
		#if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
		, isHio(false)
		#endif
	{}
	~WiiUWebInspectorServerAcceptEvent() {}
	int connectionFd;

	#if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
	int isHio;
	#endif

    virtual WiiUEventType_t type() const    { return WIIU_INSPECTOR_SERVER_ACCEPT; }
};

// Event for our inspector frontend connecting
class WiiUInspectorFrontendConnectedEvent : public WiiUBaseEventData
{
public:
	virtual WiiUEventType_t	type() const	{ return WIIU_INSPECTOR_FRONTEND_CONNECTED; }
};

// Event for dashboard reconnecting
class WiiUHostReconnectedEvent : public WiiUBaseEventData
{
public:
	virtual WiiUEventType_t	type() const	{ return WIIU_HOST_RECONNECTED; }
};

// Event for custom fonts being loaded over running client
class WiiUCustomFontsLoaded : public WiiUBaseEventData
{
public:
	virtual WiiUEventType_t	type() const	{ return WIIU_CUSTOM_FONTS_LOADED; }
};

// Event for checking if any event urls can be loaded
class WiiUTriggerEventUrlsEvent : public WiiUBaseEventData
{
public:
	virtual WiiUEventType_t	type() const	{ return WIIU_TRIGGER_EVENT_URLS; }
};

/* CRANK: DLM: This event is generated after a vsync and is used by the WebView to trigger
   any scheduled frame callbacks through requestAnimationFrame.  Since the
   runScheduledAnimations call ends up posting a task, this isn't strictly
   necessary, but it provides a more consistent interface and this event could
   be useful for other purposes in the future. */

// ----------------------------------------------------------------------------------

class WiiUDisplayVsyncEvent : public WiiUBaseEventData
{
public:
	WebViewHandle_t		webview;
	#if !PLATFORM(WIIU)
	bool m_RefreshOnly;
	#endif

	WiiUDisplayVsyncEvent(WebViewHandle_t _webview) : webview(_webview)
	{
		#if !PLATFORM(WIIU)
		m_RefreshOnly = false;
		#endif
	}

	#if !PLATFORM(WIIU)
	WiiUDisplayVsyncEvent(WebViewHandle_t _webview, bool vRefreshOnly) : webview(_webview), m_RefreshOnly(vRefreshOnly){}
	#endif

	virtual WiiUEventType_t	type() const	{ return WIIU_DISPLAY_VSYNC; }

	virtual bool rejectExclusive() const;
	virtual void postExclusive();
	virtual void releaseExclusive();

	friend void ResetWkEventGlobals();
	static void Startup(void);
private:

	static bool		s_bInQueue[WIIU_MAX_DISPLAY];
};

// Event for Remote Web Inspector Server start
class WiiUWebInspectorServerStartEvent : public WiiUBaseEventData
{
public:
    WiiUWebInspectorServerStartEvent(int port, std::string & frontendurl, std::string & hosturl)
        : listenPort(port)
        , frontendUrl(frontendurl)
        , hostUrl(hosturl){}
    ~WiiUWebInspectorServerStartEvent() {}
    int listenPort;
    std::string frontendUrl;
    std::string hostUrl;

    virtual WiiUEventType_t type() const    { return WIIU_INSPECTOR_SERVER_START; }
};

// Event for Remote Web Inspector Server stop
class WiiUWebInspectorServerStopEvent : public WiiUBaseEventData
{
public:
	virtual WiiUEventType_t type() const    { return WIIU_INSPECTOR_SERVER_STOP; }
};

// Event for Remote Web Inspector Enable
class WiiUWebInspectorEnableEvent : public WiiUBaseEventData
{
public:
    WiiUWebInspectorEnableEvent(WebView* view, bool enable)
        : inspectedView(view),
        inspectorEnable(enable) {}
    ~WiiUWebInspectorEnableEvent() {}
    WebView * inspectedView;
    int inspectorEnable;

    virtual WiiUEventType_t type() const    { return WIIU_INSPECTOR_ENABLE; }
};

class WiiURestartWebviewsEvent : public WiiUBaseEventData
{
public:
	WiiURestartWebviewsEvent(const char * vXMLFile = 0, const unsigned int vFileSize = 0, float fade = 1000.0f) : m_pConfigData(0), m_DataSize(0), m_Fade(fade)
	{
		//If we were given XML data in this event, copy it so we can update everything on reload
		if (vXMLFile && vFileSize)
		{
			m_pConfigData = new char[vFileSize];
			memcpy(m_pConfigData, vXMLFile, vFileSize);
			m_DataSize = vFileSize;
		}
	}

	~WiiURestartWebviewsEvent()
	{
		if (m_pConfigData)
			delete [] m_pConfigData;
	}

	virtual WiiUEventType_t type() const    { return WIIU_RESTART_WEBVIEWS; }

	char * m_pConfigData;
	unsigned int m_DataSize;
	float m_Fade;
};

class WiiUBackgroundNotification : public WiiUBaseEventData
{
public:
	WiiUBackgroundNotification() {}
	~WiiUBackgroundNotification() {}
	
	virtual WiiUEventType_t	type() const	{ return WIIU_BG_NOTIFICATION; }
};

class WiiUForegroundNotification : public WiiUBaseEventData
{
public:
	WiiUForegroundNotification() {}
	~WiiUForegroundNotification() {}
	
	virtual WiiUEventType_t	type() const	{ return WIIU_FG_NOTIFICATION; }
};

class WiiUShutdownNotification : public WiiUBaseEventData
{
public:
	WiiUShutdownNotification() {}
	~WiiUShutdownNotification() {}

	virtual WiiUEventType_t	type() const	{ return WIIU_SHUTDOWN_NOTIFICATION; }
};

class WiiUPauseMedia : public WiiUBaseEventData {
public:
	WiiUPauseMedia(bool shouldPause) : m_shouldPause(shouldPause) {}

	virtual WiiUEventType_t	type() const	{ return WIIU_PAUSE_MEDIA; }
	bool m_shouldPause;
};

class WiiULightGC : public WiiUBaseEventData {
public:
	WiiULightGC(int level, bool pruneObjectCache) : m_level(level), m_pruneObjectCache(pruneObjectCache) {}
	virtual WiiUEventType_t type(void) const {return WIIU_LIGHT_GC;}
	int m_level;
	bool m_pruneObjectCache;
};

class WiiUForceGC : public WiiUBaseEventData {
public:
	WiiUForceGC(pthread_cond_t * vSignal=NULL, pthread_mutex_t * vMutex=NULL) : m_pSignal(vSignal), m_pMutex(vMutex){}
	virtual WiiUEventType_t type(void) const {return WIIU_FORCE_GC;}
	pthread_cond_t * m_pSignal;
	pthread_mutex_t * m_pMutex;
};

class WiiUStorageSyncEvent : public WiiUBaseEventData
{
public:
	WiiUStorageSyncEvent(RefPtr<WebCore::StorageSyncCallback> &storageSyncCallback, int numItemsSynced = 0, int numItemsRemaining = 0)
		: m_storageSyncCallback(&storageSyncCallback), m_numItemsSynced(numItemsSynced), m_numItemsRemaining(numItemsRemaining){}
	~WiiUStorageSyncEvent() {}
	
	RefPtr<WebCore::StorageSyncCallback>	*m_storageSyncCallback;
	int										m_numItemsSynced;
	int										m_numItemsRemaining;

	virtual WiiUEventType_t	type() const	{ return WIIU_STORAGE_SYNC; }
};

class WiiUHIONewConnectionEvent : public WiiUBaseEventData
{
public:
	WiiUHIONewConnectionEvent(){}
	virtual WiiUEventType_t	type() const { return WIIU_HIO_NEW_CONNECTION; }

};

class WiiUHIONoConnectionsEvent : public WiiUBaseEventData
{
public:
	WiiUHIONoConnectionsEvent(){}
	virtual WiiUEventType_t	type() const { return WIIU_HIO_NO_CONNECTIONS; }

};

// ------------------------------------------------------------------------------------------------------

void wk_InitEvents();
void wk_ExitEvents();

#endif
