/*

	LibShim Startup/Shutdown/Utilities

	Include

*/

#ifndef _SHIM_H_
#define _SHIM_H_

#ifdef __cplusplus
extern "C" {
#endif

extern void shim_Init(void);
extern void shim_Exit(signed forced);

void shim_FreeThreadKeyArray(void);
extern void shim_ExitWeakThreads(void);

extern void shim_SetStorageWaitReadyCallback(int (*WaitForStorageReady)());

extern void* shim_GetFSClient(void);
extern int shim_GetFSMaxArgPathSize();
extern int shim_GetFSMaxFullPathSize();
extern int shim_GetFSMaxEntNameSize();
extern const char* shim_GetFSContentDir();
extern const char* shim_GetFSSaveAccountDir();
extern const char* shim_GetFSSaveCommonDir();
// NOTE: shim_GetFSAOCDir is not thread safe and the string that is returned should be immediately
//       copied for later use since the string is overwritten each time this function is called
extern const char* shim_GetFSAOCDir(unsigned long long titleId);
extern bool shim_GetUseSaveCommon();
extern bool shim_GetUseSaveAccount();
extern void shim_SetUseSaveCommon(bool bUseSaveCommon);
extern void shim_SetUseSaveAccount(bool bUseSaveAccount);

extern const char* shim_GetConfigXMLFile();
extern int shim_GetConfigXMLSize();
extern void shim_SetDefaultConfigXML(const char* fileName);
extern void shim_SetConfigXMLFile(const char *xmlData, int size);
extern void shim_FlushStorageOnExit(bool bFlushStorageOnExit);
extern void shim_FlushPendingFileOperations();
extern long long shim_TimeOfLastFileOperation(void);
extern long long shim_TimeOfLastStorageFlush(void);

// If the font data size is 0, then font data is just a string containing the font file path
//  Otherwise, the font data returned is the raw font data
extern const char* shim_GetFontData(const char* fontFile);
extern long long shim_GetFontDataSize(const char* fontFile);
extern long long shim_GetFontModificationTime(const char* fontFile);
extern void shim_SetFontData(const char *fontFile, const char *fontData, long long fontDataSize = 0, long long fontModificationTime = 0);
extern void shim_ClearFontData();

#define SHIM_FS_CLIENT_MAX_NAME		32
typedef struct SHIMfsClient_t
	{
	void	*mpFSClient;
	char	maClientName[SHIM_FS_CLIENT_MAX_NAME];
	} SHIMfsClient;

extern int shim_GetAllFSClients(SHIMfsClient *paOutClients, int maxOutClients);

#define SHIM_FS_FILE_MAX_NAME		512

typedef struct SHIMfsErrorInfo_t {
		int		m_fsErrorCode;
		int		m_dispErrorCode;
		char	m_fileName[SHIM_FS_FILE_MAX_NAME];
} SHIMfsErrorInfo;

#define SHIM_SO_HOST_MAX_NAME		512

typedef struct SHIMsoErrorInfo_t {
		int		m_soErrorCode;
		int		m_dispErrorCode;
		char	m_hostName[SHIM_SO_HOST_MAX_NAME];
} SHIMsoErrorInfo;

extern SHIMfsErrorInfo *shim_GetFSError(void);
extern int shim_HasFSError(void);

extern SHIMsoErrorInfo *shim_GetSOError(void);
extern int shim_HasSOError(void);

extern const char *shim_GetLanguage(void);

extern void shim_ClearFileCache(void);

#ifdef __cplusplus
}
#endif

#endif // _SHIM_H_




