/*---------------------------------------------------------------------------*

  Copyright 2010-2013 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/
// -----------------------------------------------------------------------------
//  wkaudAudioDevice.h
//
// -----------------------------------------------------------------------------

#ifndef __WKAUDAUDIODEVICE_H__
#define __WKAUDAUDIODEVICE_H__

#include <wkaudio/wkaudAudio.h>
#include <wkaudio/wkaudAudioPlatform.h>
#include <wkaudio/wkaudAudioNode.h>

#include "AudioBuffer.h"
#include "AudioNode.h"
#include "AudioParam.h"
#include <vector>
#include <queue>
#include <map>

#define MAX_PLAYING_VOICES_DEFAULT	128
#define MAX_PLAYING_VOICES_LOWEST	128
#define MAX_PLAYING_VOICES_HIGHEST	512

typedef void (*AudioDeviceNodeInputCallback)( WebCore::AudioNodeInput** pNodeInput, WebCore::AudioNode* pNode, unsigned int nInput );
typedef void (*AudioDeviceInputConnectCallback)( WebCore::AudioNode** pNodeConnect, WebCore::AudioNodeInput* pNodeInput, unsigned int nConnection );
typedef void (*AudioDeviceTypeIdentifyCallback)( unsigned* pDeviceType, WebCore::AudioNode* pNode );
typedef void (*AudioDeviceGetParamValueCallback)( float* pValue, WebCore::AudioParam* pParam );
typedef void (*AudioDeviceGetBufferDataCallback)( wkAudio::axAudioBuffer* axBuffer );
typedef void (*AudioDeviceStopPlaybackCallback)( WebCore::AudioNode* pNode );
typedef void (*AudioDeviceWriteAnalyserCallback)( WebCore::AudioNode* pNode, s32* pData, u32 nChannels, u32 nSamples );
typedef void (*AudioDeviceGetPannerDataCallback)( WebCore::AudioNode* pNode, float* pAzimuth, float* pElevation, float* pGain, float* pPitch );
typedef void (*AudioDeviceDoConvolverCallback)( WebCore::AudioNode* pNode, s32* pData, u32 nChannels, u32 nSamples );


using namespace std;
using namespace WebCore;

namespace wkAudio
{

// Audio Device Mirror
//
class AudioDeviceMirror
{
public:
	static AudioDeviceMirror*	instance() { if (s_pInstance == NULL) s_pInstance = new AudioDeviceMirror(); return s_pInstance; };

	static void destroy() { if (s_pInstance) { delete s_pInstance;  s_pInstance = NULL; } };

	void InitializeRemoteCallbacks(
			AudioDeviceNodeInputCallback		nodeInput_callback = 0,
			AudioDeviceInputConnectCallback		inputConnect_callback = 0,
			AudioDeviceTypeIdentifyCallback		typeIdentify_callback = 0,
			AudioDeviceGetParamValueCallback	getParamValue_callback = 0,
			AudioDeviceGetBufferDataCallback	getBufferData_callback = 0,
			AudioDeviceStopPlaybackCallback		stopPlayback_callback = 0,
			AudioDeviceWriteAnalyserCallback	writeAnalyser_callback = 0,
			AudioDeviceGetPannerDataCallback	getPannerData_callback = 0,
			AudioDeviceDoConvolverCallback		doConvolver_callback = 0
		);

	// call this function from the webcore destination node
	void audioMirror( AudioNode* pDestination );

	// call these methods to keep the audio mirror informed of changes in the node graph.
	void nodeSourceDataChanged(		AudioNode* pSource );													// call when node data values are being changed
	void nodeSourceCreate(			AudioContext* pContext, AudioNode* pSource );							// call when a node is created
	void nodeSourceDestroy(			AudioContext* pContext, AudioNode* pSource );							// call when a node is destroyed
	void nodeSourceConnectTo(		AudioContext* pContext, AudioNode* pSource, AudioNode* pDestination );	// call when a node is connected to an existing node
	void nodeSourceDisconnectFrom(	AudioContext* pContext, AudioNode* pSource, AudioNode* pDestination );	// call when a node is disconnected from an existing node

	void audioMirrorStartup();
	void audioMirrorShutdown();
	void postProcessCleanup();
	void lostAudioContext();							// call this when we lose the audio context

	int  playingVoices()					{ return m_nPlayingVoices; };		// number of voices currently playing
	int	 maxPlayingVoices()					{ return m_maxPlayingVoices; };		// maximum simultaneous playing voices
	void setMaxPlayingVoices( int nVoices )	{ m_maxPlayingVoices = nVoices < MAX_PLAYING_VOICES_LOWEST ? MAX_PLAYING_VOICES_LOWEST
											: (nVoices > MAX_PLAYING_VOICES_HIGHEST ? MAX_PLAYING_VOICES_HIGHEST : nVoices); };
	void decrementPlayingVoices()			{ if (m_nPlayingVoices > 0) --m_nPlayingVoices; };

	// accessors for the audio frame callback and voice callback routines to read voice parameters
	axAudioNode*	axNodes_AFC_Next( bool getFirstNode );	// return the first/next audio node in m_axNodes, or NULL if there is none
	axAudioNode*	axNodes_VC_Next(  bool getFirstNode );	// return the first/next audio node in m_axNodes, or NULL if there is none

private:
	AudioDeviceMirror();
	virtual ~AudioDeviceMirror();

	axAudioBuffer* findBuffer(   AudioBuffer* pBuff );	// find an existing axAudioBuffer, return 0 if not found
	axAudioBuffer* getBuffer(    AudioBuffer* pBuff );	// get an existing axAudioBuffer, or create a new one
	axAudioBuffer* createBuffer( AudioBuffer* pBuff );	// create a new axAudioBuffer
	void removeBuffer( axAudioBuffer* axBuffer );		// remove an existing axAudioBuffer

	axAudioNode* findNode( AudioNode* pNode );			// find an existing axAudioNode, return 0 if not found
	axAudioNode* createNode( AudioNode* pNode );		// create a new axAudioNode
	void removeNode( axAudioNode* axNode );				// remove an existing axAudioNode

	s16 s16Volume(	float floatVolume );				// convert float volume to signed int (log scale)
	s16 s16PanLR( float floatPanLR, signed channel );	// convert float panLR to signed int
	s16 s16PanFB( float floatPanFB, signed channel );	// convert float panFB to signed int

	void calculatePan(      axAudioNode* axSource );	// calculate panLR and panFB from source model
	void setVoiceVolume(	axAudioNode* axSource, unsigned channel );
	void adjustVoiceVolume(	axAudioNode* axSource );
	void updateAuxChannels( axAudioNode* axSource, unsigned channel );
	void updateCompressor(  axAudioNode* axSource );
	void updateFilter(      axAudioNode* axSource );
	bool acquireNewVoice(	axAudioNode* axSource, unsigned channel );
	void updateAXVoices();

	unsigned assignAuxEffect( unsigned device, unsigned effect );	// assign an Aux effect to an available channel

	bool connectNodeAudioOutputDevice(	axAudioNode* axSource, axAudioNode* axDest );
	bool connectNodeAudioBufferSource(	axAudioNode* axSource, axAudioNode* axDest );
	bool connectNodeGain(				axAudioNode* axSource, axAudioNode* axDest );
	bool connectNodePanner(				axAudioNode* axSource, axAudioNode* axDest );
	bool connectNodeConvolver(			axAudioNode* axSource, axAudioNode* axDest );
	bool connectNodeDelay(				axAudioNode* axSource, axAudioNode* axDest );
	bool connectNodeChorus(				axAudioNode* axSource, axAudioNode* axDest );
	bool connectNodeReverb(				axAudioNode* axSource, axAudioNode* axDest );
	bool connectDynamicsCompressor(		axAudioNode* axSource, axAudioNode* axDest );
	bool connectNodeAnalyser(			axAudioNode* axSource, axAudioNode* axDest );
	bool connectBiQuadFilter(			axAudioNode* axSource, axAudioNode* axDest );
	bool connectLowPass2Filter(			axAudioNode* axSource, axAudioNode* axDest );
	bool connectHighPass2Filter(		axAudioNode* axSource, axAudioNode* axDest );

	bool connectNodes( axAudioNode* axSource, axAudioNode* axDest );

	void scanNodeInputs( axAudioNode* axNode );
	void traverseGraph(  axAudioNode* axDest );

	void createAndDestroyNodes();

	void resetAudioParameters();
	void auxEffectsStartup();
	void auxEffectsShutdown();
	void registerAudioCallback();
	void deregisterAudioCallback();
	void axNodes_Reset_Iterators();						// reset the m_axNodes iterators

	double						m_currentTime;			// current time for this traversal
	int							m_nPlayingVoices;		// current number of playing voices
	int							m_maxPlayingVoices;		// maximum number of simultaneous playing voices

	map<AudioNode*,axAudioNode*>::iterator	m_axNodes_AFC_iterator;
	map<AudioNode*,axAudioNode*>::iterator	m_axNodes_VC_iterator;

	queue<AudioNode*>			m_axDataChangeQueue;	// node data change message queue
	queue<nodeConnectMessage*>	m_axNodeConnectQueue;	// node connect message queue
	queue<axAudioNode*>			m_axScanQueue;			// graph building queue, scanning edge
	queue<axAudioNode*>			m_axTraverseQueue;		// graph building queue, traversal edge

	map<AudioNode*,axAudioNode*>		m_axNodes;		// map of axAudioNodes, indexed by AudioNode
	map<AudioBuffer*,axAudioBuffer*>	m_axBuffers;	// map of axAudioBuffers, indexed by AudioBuffer

	static AudioDeviceMirror*	s_pInstance;			// singleton instance
};

}; // WebCore

#endif /// __WKAUDAUDIODEVICE_H__

